/***********************************************************/
/*                  Graph Paritioning                      */
/*                       args.c                            */
/*                                                         */
/*                Author:  James P. Fox                    */
/*                Advisor: Mark Goldberg                   */
/*                                                         */
/*                    Masters Project                      */
/*            Rensselaer Polytechnic Institute             */
/*                  Summer, Fall, 1990                     */
/***********************************************************/
/*     This module provides the routines to verify and     */
/*  process the command line switches used by the graph    */
/*  partitioning programs.  In the event of the use of     */
/*  invalid options, an error is printed and a list of all */
/*  valid options is displayed for the user.               */
/***********************************************************/
/*  Valid options are:                                     */
/*     -a  maximize                                        */
/*     -i  minimize                                        */
/*     -s  short output (only best partitioning)           */
/*     -h  skip postprocessing, run by time, not iterations*/
/*     -S  print the stabilizers of nongreedy vertices	   */
/*     -d  detailed output (including each vertex step     */
/*     -D  output s.n.g. distribution file                 */
/*         with configuration info also                    */
/*     -D2 use as intervals k instead of k/total           */
/*     -V  compute variances or run PG_VG alg              */
/*     -V2 run PG_VG2 alg (modifies ng_distrib)            */
/*     -P  perform near-greedy PG heuristic (bk only)      */
/*     -K  run kl_start-pg_finish algorithm (bk only)	   */
/*     -N  no reordering before selection   (bk only)      */
/*     -A  output abs_diff distribution (abs.dist)         */
/*     -j  include in detail list of adjacent vertices     */
/*     -g  graph greedy/nongreedy vertices                 */
/*     -n  analyze P_n(z), P_g(z)                          */
/*     -gp output graphpack information                    */
/*     -gh output greedy history (gh.out)		   */
/*     -u  compute info r.e. nongreedy up-neighbors        */
/*     -G  generate geometric graphs     		   */
/*     -R  generate regular graphs 	    		   */
/*     -ur generate unbalanced random graphs               */
/*     -ug generate unbalanced regular graphs     	   */
/*     -L  <num> add <num> leaves to each vertex	   */
/*     -f  specify input file                              */
/*     -o  specify output file                             */
/*     -e  specify external edge file                      */
/*     -t  specify summary file                            */
/*     -p  specify output plot file                        */
/*     -B  compare P_n(z) with P_n'(z)  (not P_g(z))       */
/*     -v  view information about each KL pass             */
/*                                                         */
/*  Detailed explanation of these options are given in the */
/*     users guide which accompanies this program.         */
/***********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "part.h"
#include "data_sets.h"

/*char * strcpy(char * char *);*/

void read_peinado(char *);

extern int read_file_mode, mode, alg_type, p_graph; extern int Exhaustive_Search;				/** for bk, not kl   **/
extern int Augment_With_Local;
extern int Output_Greedy_History;
extern FILE *infile,*outfile,*tfile,*pfile,*efile;
extern char efilename[20];
extern int output_short, output_detail, output_adjacent ;
extern int Near_Greedy;		/** TRUE if near-gr used somehow (pg or othr)*/
extern int P_Heuristic;		/** use the table ng_distrib[][]	    **/
				/** to determine gr/nongr. steps  (bk only) **/
extern int No_BK_Reorder;	/** select vertices in bk simply by vertex  **/
				/** number, not max-diff,max-degree	    **/
int Output_P_Count;		/** put results of pg runs in cr/distr      **/
int Output_Stabilizers;		/** vertices which help increase a ng's     **/
				/** abs. diff back to 0.		    **/
int P_N_Prime;			/** flag: compare P_n(z) vs. P_n'(z) ??     **/
int Output_Gnuplot;		/** Gnuplot   output 			    **/
int Output_GW;			/** Goemans/Williamson output		    **/
int Output_GP;			/** output	Graphpack input format	    **/
int Output_Chaco;		/** output	Chaco input format	    **/
int Output_Rao;			/** output	Rao input format            **/
int Output_Leda;		/** output 	Leda input format	    **/
int Nongreedy_Stats;		/** call nongreedy_stats()		    **/
int Find_Clusters; 		/** of nongreedy steps.   		    **/
double Cluster_Definition;	/** act int < c * ave_interval ==> cluster  **/
int Abs_Distfile;		/** create abs.dist			    **/
int View_Pass;			/** look at the pos & neg moves by KL	    **/
int Peinado_Graphs;		/** use      these instead of  (n,p) graphs **/
int Geometric_Graphs;		/** generate these instead of  (n,p) graphs **/
int Regular_Graphs;		/** generate these instead of  (n,p) graphs **/
int Unbalanced_Regular_Graphs;	/** generate these instead of  (n,p) graphs **/
int Unbalanced_Random_Graphs;	/** generate these instead of  (n,p) graphs **/
int Simulated_Annealing;	/** run this instead of KL            	    **/
int Heuristic_Only;		/** surpress postprocessing/no checks/	    **/
				/** and run for <cpu_seconds>, not <iter>   **/
int NG_Upflag;			/** Compute info r.e. nongreedy "upneighbors"*/
				/** of a vertex being placed		    **/
int Leaves;			/** allow k leaves adjacent to each vert.   **/
int Up_Down_KL;			/** Run KL with some subset of vertices     **/
				/** fixed.  That subset is determined by    **/
				/** the postprocessing order.		    **/
int Up_Down_Reps;		/** how many times per kl_pass do we fix &  **/
				/** complete?				    **/

int NG_10_Intervals;		/** put in distr. file intervals based on   **/
				/** ng_count of first 10:  (0, 1, 2, ..)    **/
int Abs_Diff_Intervals;		/** put in distr. file intervals based on   **/
				/** abs_diff (0, 1, 2, ...)		    **/
int Up_NG_Intervals;		/** put in distr. file intervals based on   **/
				/** # up ng neighbors (0, 1, 2, ...)        **/
int Output_Configuration_Table;	/** map of g/ng  01010100 at end of dst file**/
int Postproc_With_Mindiff;	/** min-diff/min-deg for next concurrent    **/
				/** reordering during maximization	    **/
int Many_Cands;			/** use max-diff (or min-diff), disregard   **/
				/** degree				    **/
int Use_One_Distribution;	/** might be using a kd_tab.c with many     **/
				/** distributions - it's assumed that the   **/
				/** last column is the "standard" ng prob.  **/
				/** use this last column.		    **/
int Use_Fixed_Prob;		/** 0.05 ng prob for all steps		    **/
int Use_Configurations;		/** use above map in PG operation           **/
int Use_Standard_For_Middle;	/** for use with many distributions (ng_up) **/
				/** when k/total is below the lower (ng)    **/
				/** threshold, that prob. is used, when it  **/
				/** is above the upper (gr) threshold, the  **/
				/** appropriate prob. is used.  When k/total**/
				/** falls in between these figures, instead **/
				/** of using the corresponding prob, use the**/
				/** "standard" ng prob - the prob if all    **/
				/** intervals are considered together       **/
int Path_KL;			/** KL pass uses path finding 		    **/
int Produce_Random_Solution;	/** stop before running KL                  **/
int Line_Starts;		/** Geometric Graphs - better start meth.   **/
int Dont_Compute_Costs;		/** for augmenting path optimization        **/
int W_Starts;			/** for sim. annealing                      **/
int Augment_With_Paths;		/** alternating (augmenting) paths are      **/
				/** swapped				    **/
int Long_Augmenting_Paths;	/** find them by tracing adj_list, not 	    **/
				/** appropriate bucket lists                **/
int Strict_Trees;		/** if true, paths are true paths.  If not, **/
				/** any string is ok.			    **/
int Path_Size;			/** for kl only - length of paths to search **/
              			/** for.                                    **/
int Increasing_Density_Experiments;
int Increasing_Size_Experiments;
int Output_Sng_Distfile;	/** create D(i) distribution file           **/
int Compute_Variance;		/** see data_sets.c                         **/
int VG_Flag;			/** W using variance (not ng distribution)  **/
int PG_VG_Flag;			/** PG runs using variance restart bad runs **/
int PG_VG_Flag2;		/** PG runs using variance to mod. distr.   **/
int KL_PG_Flag;			/** kl fixed start, pg finish (find thresh.)**/
int KL_PG_Alg;			/** kl fixed start, pg finish (run alg.)    **/
int KL_PG_Threshold;		/** theshold to use when running above alg. **/
VG_PARM Data_Sets[NUM_DATA_SETS];/** for Variance_stats, 0-ngcount, 1totcost*/

FILE *lfile;			/** output leda input file		    **/
FILE *cfile;			/** output chaco input file		    **/
FILE *rfile;			/** output rao input file		    **/
FILE *rcfile;			/** output rao coodinates file (geom graphs)**/
/**********************************************************/
/*****************  process parameters  *******************/
/**********************************************************/
void process_parameters(int argc, char *argv[]) 
{  char *s;
   int i, e_exit;
   int index;
   read_file_mode = FALSE;
   p_graph = FALSE;
   output_short = output_detail = output_adjacent = FALSE ;
   infile = stdin;
   outfile = stdout;
   tfile = NULL;
   e_exit = FALSE;

   Output_P_Count = 0;	
   P_Heuristic = 0;	
   Near_Greedy = 0;	
   No_BK_Reorder = 0;	
   Output_Stabilizers = 0;
   Path_KL = 0;
   Dont_Compute_Costs = 0;
   W_Starts = 0;
   Augment_With_Paths = 0;
   Long_Augmenting_Paths = 0;
   Line_Starts = 0;
   Produce_Random_Solution = 0;
   Strict_Trees = 0;
   Path_Size = 0;
   Output_Greedy_History = 0;
   P_N_Prime = 0;
   Output_Gnuplot = 0;
   Output_GP = 0;
   Output_Chaco = 0;
   Output_Rao = 0;
   Output_GW = 0;
   Nongreedy_Stats = 0;
   Find_Clusters = 0;
   Abs_Distfile = 0;
   View_Pass = 0;
   Geometric_Graphs = 0;
   Peinado_Graphs = 0;
   Up_Down_KL = 0;
   Up_Down_Reps = 0;
   Regular_Graphs = 0;
   Postproc_With_Mindiff = 0;
   Many_Cands = 0;
   Use_One_Distribution = 0;
   Use_Configurations = 0;
   Use_Fixed_Prob = 0;
   Use_Standard_For_Middle = 0;
   Unbalanced_Regular_Graphs = 0;
   Unbalanced_Random_Graphs = 0;
   Simulated_Annealing = 0;
   Heuristic_Only = 0;
   Increasing_Density_Experiments = 0;
   Increasing_Size_Experiments = 0;
   Output_Sng_Distfile = 0;
   NG_10_Intervals = 0;
   Abs_Diff_Intervals = 0;
   Up_NG_Intervals = 0;
   Output_Configuration_Table = 0;
   Compute_Variance = 0;
   VG_Flag = 0;
   PG_VG_Flag = 0;
   PG_VG_Flag2 = 0;
   KL_PG_Flag = 0;
   KL_PG_Alg = 0;
   KL_PG_Threshold = 0;
   Output_Leda = 0;
   Leaves = 0;
   for (i=0;i<NUM_DATA_SETS; i++)
	Data_Sets[i].reverse = -1;	/* a flag to variance_stats_ok() */

   if(argc <2)
      e_exit = TRUE;
   while(--argc > 0 && (*++argv)[0] == '-') {
     /*for(s=argv[0]+1; *s != '\0'&& !e_exit; s++)*/
     s=argv[0]+1;
	switch(*s) {
	   case 'q' :  /** minimum quotient cut **/
		       if (Simulated_Annealing)
				mode = MODE_MIN;   /* best way is to min. cut */
		       else			   /* then balance part.      */
		       		mode = MIN_QUOTIENT;
		       break;
	   case 'i' :  /** minimize **/
		       mode = MODE_MIN;
		       break;
	   case 'a' :  /**  maximize **/
		       mode = MODE_MAX;
		       break;
	   case 'E' :  /**  exhaustive backtracking search **/
		       Exhaustive_Search = TRUE;
		       break;
	   case 'h' :  /** Heuristic only **/
		       Heuristic_Only = TRUE;
		       break;
	   case 's' :  /** short output */
		       output_short = TRUE;
		       break;
	   case 'b' :  /**  print stabilizers of nongreedy vertices **/
		       Output_Stabilizers = TRUE;
		       break;
	   case 'd' :  /**  detailed output  **/
	               if (*(s+1) == 'u') {
				Output_Gnuplot = TRUE;
				break;
		       } else {
		       		output_detail = TRUE;
		       		break;
		       }
	   case 'P' :  /**  if bk, use near-greedy probability heuristic */
	               if (*(s+1) == 'G') {
				Use_One_Distribution = TRUE;
	               } else if (*(s+1) == 'F') {
				Use_Fixed_Prob = TRUE;
	               } else if (*(s+1) == 'O') {
				Output_P_Count = TRUE;
		       } else if (*(s+1) == 'C') {
				Use_Configurations = TRUE;
		       } else if (*(s+1) == 'S') {
				Use_Standard_For_Middle = TRUE;
		       } else if (*(s+1) == 'K') {
                       		Up_NG_Intervals = TRUE;
		       } else if (*(s+1) == 'T') {
                       		NG_10_Intervals = TRUE;
	               } else if (*(s+1) == 'A')
                       		Abs_Diff_Intervals = TRUE;
		       P_Heuristic = TRUE;
		       Near_Greedy = TRUE;
		       break;
	   case 'K' :  /**  if bk, use kl_start, pg_finish method */
		       KL_PG_Flag = TRUE;
	               if (*(s+1) == '2') {
		       		KL_PG_Alg = TRUE;
		       		--argc;
				if (!(++argv)[0]) {
				   fprintf(outfile,"Arg. Error (-K2): ");
				   fprintf(outfile,"int expected.\n");
				   exit(4940);
				} if (!sscanf(argv[0],"%d", 
		       	    	    &KL_PG_Threshold)){
				   fprintf(outfile,"Arg. Error (-K2): ");
				   fprintf(outfile,"int expected.\n");
				   exit(4491);
				}
		       }
		       break;
	   case 'N' :  /**  if bk, select in order of vertex number   */
		       No_BK_Reorder = TRUE;
		       break;
	   case 'S' :  /**  strongly nongreedy D(i) file **/
	               if (*(s+1) == 'A') {
				Simulated_Annealing = TRUE;
				if (mode == MIN_QUOTIENT)
					mode = MODE_MIN;
				break;
		       }
                       VG_Flag = TRUE;
		       Nongreedy_Stats  = TRUE;
		       if (alg_type == BACKTRACKING) {
				--argc;
				if (!(++argv)[0]) {
				   fprintf(outfile,"Arg. Error (-S): ");
				   fprintf(outfile,"double expected.\n");
				   exit(44590);
				} if (!sscanf(argv[0],"%lf", 
		       	    	    &Data_Sets[NG_COUNT].std_dev_mult)){
				   fprintf(outfile,"Arg. Error (-S) :");
				   fprintf(outfile,"double expected.\n");
				   exit(44591);
				}
		       }
		       break;
	   case 'V' :  /**  strongly nongreedy D(i) file **/
                       Near_Greedy = TRUE;
	               if (*(s+1) == '2') {
				PG_VG_Flag2 = TRUE;
		       		if (alg_type == BACKTRACKING) {
					--argc;
					if (!(++argv)[0]) {
					   fprintf(outfile,"Arg. Error(-V2): ");
					   fprintf(outfile,"double expected.\n");
					   exit(44590);
					}
					if (!sscanf(argv[0],"%lf", 
		       	    		    &Data_Sets[NG_COUNT].std_dev_mult)){
					   fprintf(outfile,"Arg. Error(-V2):");
					   fprintf(outfile,"double expected.\n");
					   exit(44591);
					}
		       		}
		       		Nongreedy_Stats  = TRUE;
				break;
		       }
                       Compute_Variance = TRUE;
                       PG_VG_Flag = TRUE;
		       Nongreedy_Stats  = TRUE;
		       if (alg_type == BACKTRACKING) {
		       		--argc;
		       		index = atoi((++argv)[0]);
		       		--argc;
		       		Data_Sets[index].std_dev_mult=atof((++argv)[0]);
		       		--argc;
		       		Data_Sets[index].reverse = atoi((++argv)[0]);
		       }
		       break;
	   case 'C' :  /**  clusters of nongreedy steps file **/
                       Find_Clusters  = TRUE;
		       Nongreedy_Stats = TRUE;
		       	--argc;
		       	Cluster_Definition=atof((++argv)[0]);
		        break;
	   case 'D' :  /**  strongly nongreedy D(i) file **/
		       Nongreedy_Stats = TRUE;
                       Output_Sng_Distfile = TRUE;
                       Output_Configuration_Table = TRUE;
	               if (*(s+1) == '2')
                       		Up_NG_Intervals = TRUE;
	               else if (*(s+1) == 'A')
                       		Abs_Diff_Intervals = TRUE;
	               else if (*(s+1) == 'T')
                       		NG_10_Intervals = TRUE;
	               if (*(s+2) == 'd')
                       		Increasing_Density_Experiments = TRUE;
	               else if (*(s+2) == 's')
                       		Increasing_Size_Experiments = TRUE;
		       break;
	   case 'A' :  /**  output data for graph of abs_diff **/
		       Abs_Distfile = TRUE;
		       break;
	   case 'r' :  Produce_Random_Solution = TRUE;	
		       break;
	   case 'u' :  if (*(s+1) == 'r') {	
				Unbalanced_Random_Graphs = TRUE;
				++s;
		       		break;
	               } else if (*(s+1) == 'g') {
				Unbalanced_Regular_Graphs = TRUE;
				++s;
		       		break;
		       } else {		/** graph g/ng **/
		       		NG_Upflag = TRUE;
		       		break;
		       }
	   case 'j' :  /**  output adjacent vertices **/
		       output_adjacent = TRUE;
		       break;
	   case 'G' :  /**  create geometric graphs instead of (n,p) **/
		       Geometric_Graphs = TRUE;
		       break;
	   case 'R' :  /**  create regular graphs instead of (n,p) **/
		       Regular_Graphs = TRUE;
		       break;
	   case 'U' :  /** use up-down KL alg **/
		       --argc;
		       Up_Down_KL = TRUE;
		       Up_Down_Reps = atoi((++argv)[0]);
		       break;
	   case 'w' :  
		       W_Starts = TRUE;
		       break;
	   case 'l' :  
	               if (*(s+1) == 'i') {	
				Line_Starts = TRUE;
				break;
		       }
	               if (*(s+1) == 'o') {	
				Augment_With_Local = TRUE;
				break;
		       }
		       Output_Leda = TRUE;
		       --argc;
		       if ((lfile = fopen((++argv)[0],"w")) == NULL) {
			  printf("### can't open LEDA file\n");
			  e_exit = TRUE;
		       } 
		       break;
	   case 'L' :  /**  add leaves to the graph **/
		       --argc;
		       Leaves = atoi((++argv)[0]);
		       break;
	   case 'g' :  if (*(s+1) == 'p') {	/** output graphpack**/
				Output_GP = TRUE;
		       		break;
	               } else if (*(s+1) == 'P') {
				--argc;
				Peinado_Graphs = TRUE;
				read_peinado((++argv)[0]);
		       		break;
	               } else if (*(s+1) == 'w') {
				Output_GW = TRUE;
		       		break;
	               } else if (*(s+1) == 'h') {
				Output_Greedy_History = TRUE;
		       		break;
		       } else {		/** graph g/ng **/
		       		p_graph = TRUE;
		       		break;
		       }
	   case 'm' :  if (strcmp(s,"min_diff") == 0)
				Postproc_With_Mindiff = TRUE;
		       else 
				Many_Cands = TRUE; 
		       break;
	   case 'k' :  /* augmenting paths, don't compute costs & gains */
		       {
				if (Augment_With_Paths) {
					fprintf(outfile,"must pick either -c");
					fprintf(outfile,"or -k\n"), exit(8952);
				}
				Augment_With_Paths = TRUE;
				Dont_Compute_Costs = TRUE;
	               		if (*(s+1) == 's') Strict_Trees = TRUE;
		       		break;
		       }
	   case 'c' :  /* augmenting paths, compute costs and gains */
		       {
				if (Augment_With_Paths) {
					fprintf(outfile,"must pick either -c");
					fprintf(outfile,"or -k\n"), exit(8952);
				}
				Augment_With_Paths = TRUE;
	               		if (*(s+1) == 'l') {
					Long_Augmenting_Paths = TRUE;
					s++;
				}
	               		if (*(s+1) == 's') Strict_Trees = TRUE;
		       		if (alg_type == ALG_TYPE_KL) {
	               			if ((*(s+1) == 'p')|| (*(s+2) == 'p')) {
					  Path_KL = TRUE;
					  --argc;
					  if (!(++argv)[0]) {
					   fprintf(outfile,"Arg. Error(-c): ");
					   fprintf(outfile,"int expected.\n");
					   exit(43592);
					  }
					  if (!sscanf(argv[0],"%d", 
		       	    		    &Path_Size)){
					   fprintf(outfile,"Arg. Error(-c):");
					   fprintf(outfile,"int expected.\n");
					   exit(43593);
				  	  }
	               			  if((*(s+2)=='n')||(*(s+3)=='n')) 
						Augment_With_Paths = FALSE;
					}
		       		}
				break;
		       }
	   case 'f' :  /** input file **/
		       --argc;
		       infile = freopen((++argv)[0],"r",stdin) ;
		       if(infile == NULL) {
			  printf("### invalid input file\n");
			  e_exit = TRUE;
			}
		       else {
			  read_file_mode = TRUE;
			}
		       break;
	   case 'e' :  /** external edge file **/
		       --argc;
		       efile = fopen((++argv)[0],"r") ;
		       if(efile == NULL) {
				printf("### invalid external edge file\n");
				e_exit = TRUE;
		       } else {
			  	strcpy(efilename,argv[0]);
		       }
		       break;
	   case 'n' :  /**  output adjacent vertices **/
		       Nongreedy_Stats = TRUE;
		       break;
	   case 'o' :  /** output file  **/
                       if (*(s+1) == 'c') {
				Output_Chaco = TRUE;
                                --argc;
		       		if ((cfile = fopen((++argv)[0],"w")) == NULL) {
                                     fprintf(outfile,"Arg. Error(-oc): ");
                                     fprintf(outfile,"filename expected.\n");
                                     exit(4390392);
                                }
				break;
		       }
                       if (*(s+1) == 'r') {
				char rfilename[200];
				char rcfilename[200];
				Output_Rao = TRUE;
                                --argc;
				sprintf(rfilename,"%s.adj", (++argv)[0]);
				sprintf(rcfilename,"%s.init", (argv)[0]);
		       		if ((rfile = fopen(rfilename,"w")) == NULL) {
                                     fprintf(outfile,"Arg. Error(-or): ");
                                     fprintf(outfile,"filename expected.\n");
                                     exit(49092);
                                }
		       		if (Line_Starts && 
				    ((rcfile=fopen(rcfilename,"w")) == NULL)) {
                                     fprintf(outfile,"Arg. Error(-or): ");
                                     fprintf(outfile,"filename expected.\n");
                                     exit(40785992);
                                }
				break;
		       }
		       --argc;
		       outfile = fopen((++argv)[0],"r") ;
		       if(outfile != NULL) {
			  printf("### output file exists\n");
			  e_exit = TRUE;
		       }
		       else {
			  outfile = fopen(argv[0],"w") ;
			  if(outfile == NULL) {
			     printf("### unable to open output file\n");
			     e_exit = TRUE;
			   }
		       }
		       break;
            case 't':  /** output summary file **/
		       --argc;
		       tfile = fopen((++argv)[0],"r");
		       if(tfile != NULL) { /** file exists, so open in append mode **/
		          fclose(tfile);
		          tfile = fopen((argv)[0],"a");
		          if(tfile ==NULL) { /** uable to append to summary file **/
			     printf("### unable to append to summary file\n");
			     e_exit = TRUE;
			   }
		          else
			     printf("*** summary file exist, new output will be appended\n");
			}
		       else {
			  tfile = fopen(argv[0],"w");
			  if(tfile == NULL) { /** unable to create summary file **/
			     printf("### unable to open summary file\n");
			     e_exit = TRUE;
			   }
			}
		      break;
	    case 'B':
			P_N_Prime = TRUE;
			break;
	    case 'p':
		      --argc;
		      pfile = fopen((++argv)[0],"w");
		      if(pfile == NULL) {  /** unable to create plot file **/
			 e_exit = TRUE;
		      }
		      break;
	    case 'v':
		      View_Pass = TRUE;
		      break;
	    default:
		      printf("illegal option %c\n",*s);
		      e_exit = TRUE;
		      break;
		     }
           if(e_exit) break;
   }

   if(output_adjacent)
      if(!output_detail) {
         printf("/07/07/07/n");
         printf("   To print adjacent vertices, you must also specify\n");
         printf("      detail option\n\n");
         e_exit = TRUE;
       }

   if(e_exit) {
      printf("\n\n");
      if(alg_type == ALG_TYPE_KL) {
         printf("***************************\n");
         printf("* Kernighan-Lin algorithm *\n");
         printf("***************************\n");
       }
      else {
         printf("**********************************\n");
         printf("* Near-Greedy Graph Partitioning *\n");
         printf("**********************************\n");
     }
      printf("=== Valid command options:\n");
      printf("     -a  maximize\n");
      printf("     -i  minimize\n");
      printf("     -q  minimum quotient cut\n");
      printf("     -s   short output\n");
      printf("     -h  no postproc, run by time, not iterations\n");
      printf("     -d  detailed output\n");
      printf("     -du dump gnuplot files for cut & noncut edges\n");
      printf("     -D  output s.n.g. distributions ");
      printf(" (intervals based on (ng upneighbors)/(total upneighbors)\n");
      printf("     -D2 output s.n.g. distributions/config info ");
      printf(" (intervals based on ng upneighbors)\n");
      printf("     -DA output s.n.g. distributions ");
      printf(" (intervals based on abs_diff)\n");
      printf("     -D?d put in directory for comparisons by incr. density\n");
      printf("     -D?s put in directory for comparisons by incr. size\n");
      printf("     -V  compute variances (~/greedy/density_functions),PG_VG\n");
      printf("     -V2 run PG with variance modification to ng-distr,PG_VG2\n");
      printf("     -S  run greedy variance algorithm (not PG) \n");
      printf("     -SA run simulated annealing instead of KL \n");
      printf("     -C  find ng clusters  (~/greedy/clusters)\n");
      printf("     -P  use PG() heuristic  (bk only)\n");
      printf("     -k  use augmenting paths optimization, no cost comput.\n");
      printf("     -ks use augmenting paths optimiz, no cost comp.(strict)\n");
      printf("     -c  use augmenting paths optimization\n");
      printf("     -cl(...) use long augmenting paths optimization\n");
      printf("     -cs use augmenting paths optimization (strict) \n");
      printf("     -cp  <int> (with kl) use augmenting paths kl\n");
      printf("     -csp <int> (with kl) use augmenting paths kl (strict) \n");
      printf("     -cpn  <int> (with kl) use augmenting paths kl (no opt) \n");
      printf("     -cspn <int> (with kl) use aug paths kl (strict)(no opt) \n");
      printf("     -m  select candidates by max-diff, ignore degree \n");
      printf("     -min_diff  postprocess select becmes min-diff, min-deg\n");
      printf("     -li use line starts for geometric graphs \n");
      printf("     -lo use standard local opt (before any other)\n");
      printf("     -PG use PG() heuristic with one ng distrib only \n");
      printf("     -PA use PG() heuristic with abs_diff distributions\n");
      printf("     -PF use PG() heuristic with fixed ng prob of .05 for all\n");
      printf("     -PC use PG() heuristic with g/ng configs (bk only)\n");
      printf("     -K  use kl_start, pg_finish (bk only) (finds threshold)\n");
      printf("     -K2 use kl_start, pg_finish (bk only) (runs algorithm)\n");
      printf("     -N  no reordering before selection   (bk only)\n");
      printf("     -j  show adjacent vertices\n");
      printf("     -g  graph greedy/nongreedy\n");
      printf("     -gp output graphpack info (kl only)\n");
      printf("     -gw output Goemans/Williamson input format\n");
      printf("     -gP <file> use Peinado Graph as input\n");
      printf("     -B  compare P_n(z) with P_n'(z)  (not P_g(z))");
      printf("     -l  output leda graph format and quit\n");
      printf("     -gh output greedy history\n");
      printf("     -u  compute info r.e. nongreedy up-neighbors\n");
      printf("     -b  print info r.e. output stabilizers\n");
      printf("     -G  generate geometric graphs\n");
      printf("     -R  generate regular graphs\n");
      printf("     -ur generate unbalanced random graphs\n");
      printf("     -ug generate unbalanced regular graphs\n");
      printf("     -L <num> generate (any type) of graphs with <num> leaves on each vertex \n");
      printf("     -f  input file\n");
      printf("     -n  analyze P_n(z), P_g(z)\n");
      printf("     -o  output file\n");
      printf("     -e  external edge file\n");
      printf("     -t  summary file\n");
      printf("     -p  output plot file\n");
      printf("     -P  compare P_n(z) with P_n'(z)  (not P_g(z))\n");
      printf("     -v  view KL pass info\n\n\n");
      exit(99);
    }
 }
