/***********************************************************/
/*                  Graph Partitioning                     */
/*                      postproc.c                         */
/*                                                         */
/*                Author:  James P. Fox                    */
/*                Advisor: Mark Goldberg                   */
/*                                                         */
/*                    Masters Project                      */
/*            Rensselaer Polytechnic Institute             */
/*                  Summer, Fall, 1990                     */
/***********************************************************/
/*     This module provides the routines to handle the     */
/*  post-processing of the partitioned graphs.  It computes*/
/*  greedy/non-greedy steps, along with the degree of      */
/*  each.  Average summary information is computed for each*/
/*  of the "different" graphs being computed.  Detail and  */
/*  summary information is also written to a separate file */
/*  when requested by the command line switches.           */
/***********************************************************/
/*  NOTE:  The fields tie1 and tie2 refer to concurrent    */
/*    reordering in the following ways:                    */
/*    tie1 - number of vertices where |a-b| are equal.     */
/*    tie2 - of the above, the number with the same degree.*/
/***********************************************************/
#include <ctype.h>
#include <stdio.h>
#include <math.h>
#include "part.h"
#include "backtrack.h"

void init_gp_info();
void gp_info(int, int, int, int, int);
int comp_partitions(int *, int *);
int search_nongreedy_list(int);
int better_degree(int,int,int);
int greedy_history(int, int, int, int *);
void init_backtrack();
double greedy_tree_walk(int mode, int paths, int print, int threshold);
void dump_cutfiles(int *, int, int);


extern int vertices[];
extern int vertex_loc[];
extern int vertex_processed[];
extern int is_ind_set[];
extern int deg[];
extern int num;
extern int mode;
extern int sort_type;
extern int look_ahead, conc_tie;
extern int output_short, output_detail, output_adjacent;
extern int node_cost[2][MAX_VERTICES];
extern int hold_ng_cost[2][MAX_VERTICES];/* the cost of placing vertex v   */
					 /* in partition p, in terms of    */
					 /* adj. with ng only		  */
extern struct adj_node *adj_list[];

extern int NG_Upflag;			/* flag for -u option	*/
extern int Paths;			/* bk walks		*/
extern int iter ;			/* kl iterations 	*/
extern int currentcost;
extern int currentminsize;
extern int alg_type;
extern int max_size;
extern int greedy, nongreedy;		/* total greedy/nongreedy steps   */
extern double min_mult;
extern double prob, time_;
extern int numedges ;
extern VLIST nongreedy_vertices[MAX_VERTS];

int Output_Greedy_History;
int even_b[MAX_VERTS];

/**************************************************************************/
/** These are statistics for which mean and variance are computed in     **/
/** data_sets.c								 **/
/**************************************************************************/
int Tot_Cost[MAX_VERTS];		/* [i] = cum cuts after placing i */
int NG_Count[MAX_VERTS];		/* [i] = cum ng   after placing i */
int Size_A[MAX_VERTS];		/* [i] = cum in A after placing i */
int Size_B[MAX_VERTS];		/* [i] = cum in B after placing i */
/**************************************************************************/

int post_tots[MAX_VERTICES];        	/* degree of greedy/nongreedy     */
int hold_num_vertices[2],
         num_vertices[2]; 		/* num of verts in each set	  */ 

int subgraph_deg[MAX_VERTS];		/* deg of vertex in subraph induced*/
					/* by the unplaced vertices.	   */

int hold_node_cost[2][MAX_VERTICES],   /* keep track of best partitioning */
    hold_future_cost[2][MAX_VERTS], /* information.                    */
    hold_vertices[MAX_VERTICES],
    hold_post_tots[MAX_VERTICES],                                             
    hold_vertex_loc[MAX_VERTICES],
    hold_even_b[MAX_VERTS],
    hold_tie_1[MAX_VERTS],  		/* conc. reord.# of ties |a-b|  */
    hold_tie_2[MAX_VERTS],  		/* conc. reord.# of tie degrees */
    hold_nongreedy,
    hold_greedy;

/***************************************** 2/22/93 **************************/
int down_greedy, down_nongreedy;	/* total greedy/nongreedy steps   */
					/* w.r.t. downneighbors		  */
int both_greedy, both_nongreedy;	/* total greedy/nongreedy steps   */
					/* w.r.t. both upneighbors and    */
					/* downneighbors		  */
int hold_down_greedy, hold_down_nongreedy;	
int hold_both_greedy, hold_both_nongreedy;
int part_diffs[MAX_VERTS];		/* for each placement, look at 	    */
					/* ALL UNPLACED VERTICES. If they   */
					/* were all			    */
					/* placed greedy at that instant    */
					/* how different is that part. from */
					/* the final partitioning.  Check   */
					/* for each vert. placement to find */
					/* out when they are the same.      */
int part_neigh_diffs[MAX_VERTS];	/* for each placement, look at 	    */
					/* DOWNNEIGHBORS.  If they were all */
					/* placed greedy at that instant    */
					/* how different is that part. from */
					/* the final partitioning.  Check   */
					/* for each vert. placement to find */
					/* out when they are the same.      */
int node_downcost[2][MAX_VERTS];	/* when v is placed, how many	    */
					/* down-neighbors does it have in   */
					/* each "pretend partition." (the   */
					/* partition that would occur if    */
					/* every unplaced vertex goes to    */
					/* its current greedy partition.    */
					/* e.g. the unplaced vert's are an  */
					/* independent set.		    */

int temp_loc[MAX_VERTS];		/* consider v's placement.  it has  */
					/* downneighbors.  temp_loc contains*/
					/* the locations of the "pretend"   */
					/* partitions of the unplaced verts */

int post_tots_down[MAX_VERTS];	/* the quantitative greediness of   */
					/* each vertex in terms of its down */
					/* neighbors and their "pretend"    */
					/* partitioning.		    */

int post_tots_both[MAX_VERTS];	/* the quantitative greediness of   */
					/* each vertex in terms of its down */
					/* neighbors and its "up" neighs.   */
int Pretend_Changes[MAX_VERTS];

int hold_node_downcost[2][MAX_VERTS];
int hold_post_tots_down[MAX_VERTS];	
int hold_post_tots_both[MAX_VERTS];	
/***************************************** 2/22/93 **************************/

double T_ng_upneighbors,			/* these to keep track of per-pass */
    T_upneighbors,			/* information			   */
    T_upneighbors_g,			/* see calc_ng_placed_neighbors()  */
    T_upneighbors_n,
    T_ng_upneighbors_g,
    T_ng_up_same_part_g,
    T_pct_ng_same_g,
    T_ng_up_other_part_g,
    T_pct_ng_other_g,     
    T_ng_upneighbors_n,
    T_ng_up_same_part_n,
    T_pct_ng_same_n,     
    T_ng_up_other_part_n, 
    T_pct_ng_other_n;

static double T_Ng_upneighbors,		/* these to keep track of summary  */
      T_Upneighbors,			/* information over all passes	   */
      T_Upneighbors_g,			/* see add_to_summary()		   */
      T_Upneighbors_n,
      T_Ng_upneighbors_g,
      T_Ng_up_same_part_g,
      T_Pct_ng_same_g,     
      T_Ng_up_other_part_g,
      T_Pct_ng_other_g,    
      T_Ng_upneighbors_n,
      T_Ng_up_same_part_n,
      T_Pct_ng_same_n,     
      T_Ng_up_other_part_n,
      T_Pct_ng_other_n;    

int abs_diff_p_n[MAX_VERTS];		/* abs_diff in P_N(z) placement   */
int abs_diff_p_g[MAX_VERTS];		/* abs_diff in P_G(z) placement   */
					/* serves doubly duty (P_n' also) */

VLIST hold_nongreedy_vertices[MAX_VERTS];/* backup list of best pass   */

/*** hold fields for summary information ***/
int Hnum, Tcount, Titerations, Tcuts, Tgreedy, Tnongreedy; 
double T_leaf_count, Tquotient;
double T_ave_cands;
int Hsort_type ;
int First ;
long Tedges ;
double Hprob, Ttime ;

int   	Tlast;
int   	Tcatch;
int   	Tsurpass;
int   	Tchanges;

int 	t_cost;			/* total cuts so far during postproc. */
int 	t_nongreedy;		/* total ng   so far during postproc. */
int 	t_part_a;		/* |A| so far during postproc	      */
int 	t_part_b;		/* |B| so far during postproc	      */

int T_NG_Count;			/* NG count for P heuristic    */
int T_NG_Before_100;		/* only when 500 vertices used */
int T_NG_Before_250;		/* only when 500 vertices used */
int T_NG_Before_400;		/* only when 500 vertices used */
int Tkl_pg_index;

int Tpart_match;		/* total of all stats x, where    */
				/* x = place where remaining      */
				/* "greedy remaining" partitioning*/
				/* matches the KL partitioning.   */
int KL_PG_Index;		/* point where kl-pg equals kl    */

double   Tchange_ave;
double   Tno_change_ave;
double	 T_Ave_Path_Length;		/* PO				*/
double 	 T_Num_Augmenting_Paths;	/* PO				*/
double 	 T_Ave_Flip_Cost;		/* PO				*/

int tie1, tie2 ;

FILE *outfile, *tfile, *pfile ;

/**********************************************************/
/***************  is vertex a greedy move   ***************/
/**********************************************************/
int is_greedy(int v,int loc)
{
	extern int t_cost;			/* keep track of # cuts     */
	extern int t_part_a;			/* keep track of # in A     */
	extern int t_part_b;			/* keep track of # in B     */

	int n,set;
	int futa,futb,costa,costb;
	double quot_a, quot_b;
	struct adj_node *point;

	costa = node_cost[SET_A][v];
	costb = node_cost[SET_B][v];
	futa = futb = 0 ;
	point = adj_list[v];
	while(point != NULL) {
		n = point->node_no ;
		if(!vertex_processed[n])
			if(look_ahead) {
				if(node_cost[SET_A][n] > node_cost[SET_B][n])
					futb++;
				else if(node_cost[SET_A][n]<node_cost[SET_B][n])
					futa++;
			}
			point = point->next;
	}
	costa += futa;
	costb += futb;
	if(mode == MODE_MIN && alg_type == ALG_TYPE_NEAR_GREEDY) {
		costa = costa + num_vertices[SET_A] * min_mult;
		costb = costb + num_vertices[SET_B] * min_mult;
	}

	if (mode == MIN_QUOTIENT) {
	   quot_a=(double)(t_cost + costa)/(double)MIN(t_part_a+1,t_part_b);
	   quot_b=(double)(t_cost + costb)/(double)MIN(t_part_a,t_part_b+1);
	   	if(quot_a>quot_b)
			set = SET_A;
		else
			set = SET_B ;
	} else {
		if(costa == costb) {
			if(loc == SET_A)
				return(TRUE);
			else
				return(FALSE);
		} 
		if(costa>costb)
			set = SET_A;
		else
			set = SET_B ;
	}
	if(mode == MODE_MAX) {
		if(set == loc)
			return(TRUE);
		else
			return(FALSE);
	} else {
		if(set == loc)
			return(FALSE);
		else
			return(TRUE);
	}
}

void init_ng_placed_neighbors()
{
	int i;

	T_ng_upneighbors     = 0.0;
	T_upneighbors        = 0.0;
	T_upneighbors_g      = 0.0;
	T_upneighbors_n      = 0.0;
	T_ng_upneighbors_g   = 0.0;
	T_ng_up_same_part_g  = 0.0;
	T_pct_ng_same_g      = 0.0;
	T_ng_up_other_part_g = 0.0;
	T_pct_ng_other_g     = 0.0;
	T_ng_upneighbors_n   =0.0;
	T_ng_up_same_part_n  = 0.0;
	T_pct_ng_same_n      = 0.0;
	T_ng_up_other_part_n = 0.0;
	T_pct_ng_other_n     = 0.0;

	for (i=0; i<num; i++) {
		hold_ng_cost[SET_A][i] = 0 ;
		hold_ng_cost[SET_B][i] = 0 ;
	}
}

void calc_ng_placed_neighbors()
/*************************************************************************/
/** calc_ng_placed_neighbors()					        **/
/**			keep running totals of some stats about	the     **/
/**			number of upneighbors of v that happen to have	**/
/**			been nongreedy steps.  The count is taken when	**/
/**			v is about to be placed.			**/
/*************************************************************************/
{
	extern int hold_post_tots[];	/* neg ==> nongreedy		   */
	extern int hold_vertices[]; 	/* vertex order from post. of best */
					/* partitioning			   */
	extern UP_INFO NG_Up_Array[];
	int i, v;
	int c_tot, ng_costa, ng_costb, node_costa, node_costb;
	int ng_upneighbors;
	double ng_pct_a, ng_pct_b, r;

	for (i=1; i<num; i++) {

	v = hold_vertices[i];
	node_costa = hold_node_cost[SET_A][v];
	node_costb = hold_node_cost[SET_B][v];
	ng_costa = hold_ng_cost[SET_A][v];
	ng_costb = hold_ng_cost[SET_B][v];
	ng_pct_a = (node_costa > 0.0) ? ng_costa / (double) node_costa : 0.0;
	ng_pct_b = (node_costb > 0.0) ? ng_costb / (double) node_costb : 0.0;

	c_tot = hold_node_cost[SET_A][v] + hold_node_cost[SET_B][v];
	r = (double) (hold_node_cost[SET_A][v] + hold_node_cost[SET_B][v]);
	ng_upneighbors   = ng_costa + ng_costb;

	T_upneighbors      += r;
	T_ng_upneighbors   += ng_upneighbors;

	/*NG_Up_Pct = (double) ng_upneighbors / (double) c_tot;*/

	if (hold_post_tots[v] >= 0  /* GREEDY */) {
		NG_Up_Array[i].up_g  += c_tot;
		NG_Up_Array[i].ng_up_g  += ng_upneighbors;
		NG_Up_Array[i].pct_ng_up_g  = 
			NG_Up_Array[i].pct_ng_up_g  +  
				((c_tot > 0) 
				   ? ((double) ng_upneighbors / (double) c_tot )
				   : 0.0);
		T_upneighbors_g      += r;
		T_ng_upneighbors_g   += ng_upneighbors;
		T_ng_up_same_part_g  += vertex_loc[v]?ng_costa:ng_costb;
		T_ng_up_other_part_g += vertex_loc[v]?ng_costb:ng_costa;
		T_pct_ng_same_g      += vertex_loc[v]? ng_pct_a : ng_pct_b;
		T_pct_ng_other_g     += vertex_loc[v]? ng_pct_b : ng_pct_a;
	} else {
		NG_Up_Array[i].times_ng  += 1;
		NG_Up_Array[i].up_n  += c_tot;
		NG_Up_Array[i].ng_up_n  += ng_upneighbors;
		NG_Up_Array[i].pct_ng_up_n  = 
			NG_Up_Array[i].pct_ng_up_n  +  
				((c_tot > 0) 
				   ? ((double) ng_upneighbors / (double) c_tot )
				   : 0.0);
		T_upneighbors_n      += r;
		T_ng_upneighbors_n   += ng_upneighbors;
		T_ng_up_same_part_n  += vertex_loc[v]?ng_costa:ng_costb;
		T_ng_up_other_part_n += vertex_loc[v]?ng_costb:ng_costa;
		T_pct_ng_same_n      += vertex_loc[v]? ng_pct_a : ng_pct_b;
		T_pct_ng_other_n     += vertex_loc[v]? ng_pct_b : ng_pct_a;
	}

	}
}

void set_part_diff(int index)
/*****************************************************************/
/** 2/21/93:  find difference between "pretend" partitioning of **/
/**           unplaced vertices based on simultaneous greedy    **/
/**	      placement,  (find this for each placement position**/
/**	      (index))  and the KL partitioning.		**/
/*****************************************************************/
{
	extern int Pretend_Changes[MAX_VERTS];
	int i, temp_loc;
	static int locs[MAX_VERTS];
	/* struct adj_node *temp; */
	/* int n,v;	          */

	if (index == 0) {
		for (i=0; i<num; i++) {
			locs[i] = 0;
			Pretend_Changes[i] = 0;
		}
	}
/*
	v = vertices[index];
	for (temp = adj_list[v]; temp!=NULL; temp=temp->next) {
		n = temp->node_no;
	        temp_loc = (node_cost[SET_B][v] > node_cost[SET_A][v]) 
			? SET_B : SET_A ; 
	     	if (temp_loc != locs[v]) {
			Pretend_Changes[v]++;
			locs[v] = temp_loc;
		}
	}
*/		

	for (i=0; i<num; i++) {
	     if (vertex_processed[i])
		temp_loc = vertex_loc[i];
	     else
	        temp_loc = (node_cost[SET_B][i] > node_cost[SET_A][i]) 
			? SET_B : SET_A ; 
	     if (temp_loc != locs[i]) {
		/*if (i == 33)
			{ int j=3; }*/
		Pretend_Changes[i]++;
		locs[i] = temp_loc;
	     }
	}
	part_diffs[index] = comp_partitions(locs, vertex_loc);
}

void process_downcost(int index, int v, int n)
/**************************************************/
/** 2/21/93:  maintain node_downcost[][]         **/
/**************************************************/
{
	int loc = (node_cost[SET_B][n] > node_cost[SET_A][n]) ? SET_B : SET_A;

	if (loc != vertex_loc[n])
		part_neigh_diffs[index]++;

	if (node_cost[SET_A][n] > node_cost[SET_B][n])
		node_downcost[SET_B][v]++;
	else
		node_downcost[SET_A][v]++;
}

/**process_downcost(int v, int n)**/
/**************************************************/
/** 2/21/93:  maintain temp_loc[i] - where it    **/
/**           would go if placed now             **/
/**	      also maintain node_downcost[][]    **/
/**************************************************/
/**{
	struct adj_node *point;
	int n2, old_loc=temp_loc[v], new_loc;

	node_downcost[!old_loc][n]--;
	if (node_downcost[!old_loc][n]>node_downcost[old_loc][n]) {
		temp_loc[n] = !old_loc;
		point = adj_list[n];
		while(point != NULL) {
			n2 = point->node_no ;
			if (!vertex_processed[n2]) {
				node_downcost[old_loc][n2]++;
				node_downcost[!old_loc][n2]--;
			}
			point = point->next;
		}
	}
}**/

/**********************************************************/
/************   get post processing node cost   ***********/
/**********************************************************/
void get_post_cost(int index, int v,int *costa,int *costb,
		   int best, int change_index)
/*  change_index:	if <change_index> = <index>, switch sides */
{  
	extern int   Done_With_Alg;
	extern int   Output_Stabilizers;
	extern VLIST nongreedy_vertices[];

	int n, p;
	int last, last2;
	int futa,futb;
	struct adj_node *point;

	*costa = node_cost[SET_A][v];
	*costb = node_cost[SET_B][v];
	futa = futb = 0 ;
	vertex_processed[v] = TRUE ;
	point = adj_list[v];
	/********************************************************************/
	/** what effect does changing the placement of only 1 vertex from  **/
	/** nongreedy to greedy have on the number of nongrdy verts found? **/
	/********************************************************************/
	if (index == change_index)
		vertex_loc[v] = !vertex_loc[v];
	else if (index > change_index) {
	   if (mode == MODE_MAX)
		vertex_loc[v] = (*costa >= *costb)? SET_A : SET_B;
	   else if (mode == MIN_QUOTIENT) 
		printf("Error: get_post_cost(): change_index w/ quotient\n");
	   else
		vertex_loc[v] = (*costa <= *costb)? SET_A : SET_B;
	}
	if (best)  set_part_diff(index);		/* 2/21/93 */
	while(point != NULL) {
		n = point->node_no ;
		if (!vertex_processed[n]) {
         		if(look_ahead) {
            			if(node_cost[SET_A][n] > node_cost[SET_B][n])
               				futb++;
            			else
               			if(node_cost[SET_A][n] < node_cost[SET_B][n])
                  			futa++;
				/*futb += node_cost[SET_A][i];*/ /* 6/27/93 */
				/*futa += node_cost[SET_B][i];*/
			}
			if (best)  process_downcost(index, v, n);  /*2/21/93*/
     			if(vertex_loc[v])
       				node_cost[SET_A][n]++;
      			else
       				node_cost[SET_B][n]++;
		} else if ((Output_Stabilizers) && 
			 (Done_With_Alg&&((p=search_nongreedy_list(n))!=EOF))){
			last =  node_cost[vertex_loc[n]][n] - 
				node_cost[!vertex_loc[n]][n];
     			if(vertex_loc[v]) 
       				node_cost[SET_A][n]++;
      			else 
       				node_cost[SET_B][n]++;
			last2 =  node_cost[vertex_loc[n]][n] - 
				node_cost[!vertex_loc[n]][n];
			if ( ((mode==MODE_MAX) && (last<0) && (last<last2)) ||
			     ((mode==MODE_MIN) && (last>0) && (last>last2)) ) {
			     if (last2 == 0)
				insert_vlist(1, index, 0, 0, 
					nongreedy_vertices[p].next);
			     else
				insert_vlist(0, index, 0, 0, 
					nongreedy_vertices[p].next);
			}
		}
      		point = point->next;
    	}
	*costa += futa;
	*costb += futb;
	if(mode == MODE_MIN && alg_type == ALG_TYPE_NEAR_GREEDY) {
		*costa = *costa + num_vertices[SET_A] * min_mult;
		*costb = *costb + num_vertices[SET_B] * min_mult;
 	}
	if(best) {
		hold_node_cost[SET_A][v] = node_cost[SET_A][v];
		hold_node_cost[SET_B][v] = node_cost[SET_B][v];
		hold_future_cost[SET_A][v] = futa;
		hold_future_cost[SET_B][v] = futb;
	}
	num_vertices[vertex_loc[v]]++;
}

int get_next_p_n_prime(int next)
{
	extern int abs_diff_p_g[];	/* used in this case for p_n' */
	int val;

	val = ABS(node_cost[SET_A][next] - node_cost[SET_B][next]);
   	First = FALSE ;
   	abs_diff_p_g[next] = val;
	return(next);
}
/**********************************************************/
/** get_next_sub_maxdeg()				 **/
/**		get next vertex for subgraph-maxdegree   **/
/** 		only used for maximization now		 **/
/**********************************************************/
int get_next_sub_maxdeg(int best_solution)
{  register j;
   int bestj, bestval, val, bestdeg;
   int n;
   struct adj_node *temp;

   if(mode == MODE_MAX) { /** maximize **/
	bestj = -1;
	bestdeg = VERY_LOW ;
	bestval = VERY_LOW ;
	for(j=0;j<num;j++) {
           if(!vertex_processed[j]) 
		if (subgraph_deg[j] > bestdeg) {
               		val = ABS(node_cost[SET_A][j] -node_cost[SET_B][j]);
			bestval = val;
			bestj = j;  
			bestdeg = subgraph_deg[j];
		}
	}
	for (temp=adj_list[bestj]; temp; temp=temp->next) {
		n = temp->node_no;
		if (!vertex_processed[n])
			subgraph_deg[n]--;
	}
	First = FALSE ;
	if (best_solution)
		abs_diff_p_n[bestj] = bestval;
	else
		abs_diff_p_g[bestj] = bestval;
	return(bestj);
   } else
	printf("ERROR: this sort type not implemented for minimization\n");
   return(0);
}

int sub_deg(int vnum)
{  
	register int n;
	int sub_degree=0;
	struct adj_node *temp;

	for (temp=adj_list[vnum]; temp; temp=temp->next) {
		n = temp->node_no;
		if (!vertex_processed[n])
			sub_degree++;
		}
	return(sub_degree);
}

int get_next_concurrent(int best_solution)
/*****************************************************************************/
/** get_next_concurrent() :	very inefficient - leads to O(n^2)          **/
/**				-h option skips postprocessing and is       **/
/**				necessary for big graphs.  Buckets could    **/
/**				efficiently do the same thing (as in bk),   **/
/**				but I haven't had time to implement them    **/
/**				here.		2/2/94			    **/
/*****************************************************************************/
{ 
  extern int Postproc_With_Mindiff;

  register j;
  int bestj,bestval,val=0;
  int diff;

  tie1 = tie2 = 0 ;
  if((mode == MODE_MAX) && !Postproc_With_Mindiff) { /** maximize **/
	bestj = -1;
	bestval = VERY_LOW ;
	for(j=0;j<num;j++) {
	     if(!vertex_processed[j] && !is_ind_set[j]) {
		if(look_ahead) { /* deleted 3/94 */
		} else
			val = ABS(node_cost[SET_A][j] - node_cost[SET_B][j]);
		if(val > bestval) {
			bestj = j; bestval = val; tie1 = tie2 = 1 ;
		} else if (val == bestval) {
			tie1++;
			if(better_degree(j,bestj,First)) {
				bestj = j; bestval = val; tie2 = 1 ;
			} else if(deg[j] == deg[bestj]) {
				tie2++;
				if(conc_tie && 
				  !is_greedy(bestj,vertex_loc[bestj])) {
					bestj = j ; bestval = val;
				}
			}
		}
	     }
	}
  } else {  /** minimize **/
	bestj = -1;
	bestval = VERY_HIGH;
	for(j=0;j<num;j++) {
  	    if(!vertex_processed[j] && !is_ind_set[j]) {
		diff = num_vertices[SET_A] - num_vertices[SET_B];
		/*diff = 0;*/
		if(look_ahead) { /* delete 3/94 */
		} else 
			val = ABS(node_cost[SET_A][j] /*+2 * diff * min_mult */ 
				 -node_cost[SET_B][j]);
		if(val < bestval) {
			bestj = j; bestval = val; tie1 = tie2 = 1 ;
		} else if (val == bestval) {
			tie1++;
			if (better_degree(j,bestj,First)) { 
				bestj = j; bestval = val; tie2 = 1 ;
			} else if(deg[j] == deg[bestj]) {
				tie2++;
				if(conc_tie && 
				  !is_greedy(bestj,vertex_loc[bestj])) {
					bestj = j ; bestval = val ;
				}
			}
		}
	    }
	}
  }
  First = FALSE ;
  if (best_solution)
	abs_diff_p_n[bestj] = bestval;
  else
	abs_diff_p_g[bestj] = bestval;
  return(bestj);
}


void delete_v(int key, VLISTPTR verts, int range)
/*****************************************************************/
/** delete_v()							**/
/**		used in storing nongreedy vertices 		**/
/*****************************************************************/
{
	int i, j;

	for (i = 0; i<range; i++)
		if (verts[i].v == key) {
			for (j=i; j<range-1; j++)
				verts[j].v = verts[j+1].v;
			break;
		}
}

void print_downcost()
/*********************** called by debugger only **********************/
{
	int i;
	printf("****node_downcost[0]: ");
	for (i=0; i<num; i++) 
		printf("%d ", node_downcost[0][i]);
	printf("\n");
	printf("****node_downcost[1]: ");
	for (i=0; i<num; i++) 
		printf("%d ", node_downcost[1][i]);
	printf("\n");
}

void find_down_greediness(int i, int v, int best)
{
	int costa = node_downcost[SET_A][v];
	int costb = node_downcost[SET_B][v];

	int amt, set;

	if (mode == MIN_QUOTIENT)
		return;

	if(costa > costb)
		amt = costa-costb;
	else
		amt = costb-costa;
	if(costa == costb) {
		down_greedy++;
		if(best)
			post_tots_down[i] = 0; 
	} else {
		if(costa>costb)
			set = SET_A;
		else 
			set = SET_B;
		if ( ((mode==MODE_MAX) && (set == vertex_loc[v])) || 
		     ((mode==MODE_MIN) && (set != vertex_loc[v])) ) { 
			down_greedy++;
			if(best)
				post_tots_down[i] = amt;
		} else {
			down_nongreedy++;
			if(best)
				post_tots_down[i] = -amt;
		}
	}
}


void find_both_greediness(int i, int v, int best)
{
	int costa = node_downcost[SET_A][v] + node_cost[SET_A][v];
	int costb = node_downcost[SET_B][v] + node_cost[SET_B][v];
	int amt, set;

	if (mode == MIN_QUOTIENT)
		return;

	if(costa > costb)
		amt = costa-costb;
	else
		amt = costb-costa;
	if(costa == costb) {
		both_greedy++;
		if(best)
			post_tots_both[i] = 0; 
	} else {
		if(costa>costb)
			set = SET_A;
		else 
			set = SET_B;
		if ( ((mode==MODE_MAX) && (set == vertex_loc[v])) || 
		     ((mode==MODE_MIN) && (set != vertex_loc[v])) ) { 
			both_greedy++;
			if(best)
				post_tots_both[i] = amt;
		} else {
			both_nongreedy++;
			if(best)
				post_tots_both[i] = -amt;
		}
	}
}

int find_up_greediness(int i, int v, int best, int costa, int costb)
{
	extern int Nongreedy_Stats;

	int greediness, amt, set, n;
	double quot_a, quot_b;
	struct adj_node *point;

	amt = ABS(costa-costb);
	if (mode == MIN_QUOTIENT) {
	   quot_a=(double)(t_cost + costa)/(double)MIN(t_part_a+1,t_part_b);
	   quot_b=(double)(t_cost + costb)/(double)MIN(t_part_a,t_part_b+1);
	   if(quot_a>quot_b)
		set = SET_A;
	   else
		set = SET_B ;
	   if (set != vertex_loc[vertices[i]]) {
		greediness = GREEDY;
		greedy++;
		if (best)   post_tots[i] = amt;
	   } else {
		greediness = NONGREEDY;
		nongreedy++;
		if (best)   post_tots[i] = -amt;
	   }
	   return(greediness);
	} 
	if(costa == costb) /*** even  ***/ {
		greediness = GREEDY;
		greedy++;
		if(best)
			post_tots[i] = 0 ; 	/** 2/21/93 changed **/
	} else {
		if(costa>costb)
			set = SET_A;
		else 
			set = SET_B;
		if ( ((mode==MODE_MAX) && (set == vertex_loc[vertices[i]])) || 
		     ((mode==MODE_MIN) && (set != vertex_loc[vertices[i]])) ) { 
			greediness = GREEDY;
			greedy++;
			if(best)
				post_tots[i] = amt; 	/** 2/21/93 **/
		} else {
			greediness = NONGREEDY;
			nongreedy++;
			if(best)
				post_tots[i] = - amt;   /** 2/21/93 **/
		}
	}
	/********************** new **************************/
	/* if(best && Output_Greedy_History &&
	   !greedy_history(i,greediness, vertices[i],vertex_processed))
		print_greedy_history(vertices[i]);*/
	/********************** new **************************/

	if (best && (NG_Upflag || Nongreedy_Stats)) {
	/*****************************************************/
	/** update down-neighbors to reflect if they have   **/
	/** a new nongreedy up-neighbor.		    **/
	/*****************************************************/
	   if (greediness == NONGREEDY) {
	   	point = adj_list[v];
	   	while(point != NULL) {
			n = point->node_no ;
			if (!vertex_processed[n])  {
				if (vertex_loc[v])
       					hold_ng_cost[SET_A][n]++;
      				else
       					hold_ng_cost[SET_B][n]++;
			}
			point = point->next;
	   	}
	   }
	}
	return(greediness);
}


/*****************************************************************/
/**  post_proc()   					 	**/
/**		Knowing the KL solution, perform a constructive **/
/**		partitioning and note the differences.		**/
/**		if this is the "best" partition so far, store	**/
/**		the results for later graphing of best parts. 	**/
/**                                                             **/
/**  parms: best - flag (is the partition the best so far?)     **/
/**         change_index - pos. of a vertex that will be        **/
/**                        switched to the other partition.     **/
/**	    duplicate_p_n - use same ordering when switching z  **/
/**	    begin_fix     - fix from here to end_fix for the    **/
/**	    end_fix       - next pass of KL or whatever alg.    **/
/*****************************************************************/
void post_proc (int best, int change_index, int duplicate_p_n, 
		int begin_fix, int end_fix)
{  
	extern int Tot_Cost[];
	extern int Output_Gnuplot;
	extern int Output_GP;			/* prepare graphpack output */
	/* extern int Output_Greedy_History;*/
	extern int Nongreedy_Stats;
	extern int sort_type;
	extern int t_cost;			/* keep track of # cuts     */
	extern int t_nongreedy;			/* keep track of # ng       */
	extern int t_part_a;			/* keep track of # in A     */
	extern int t_part_b;			/* keep track of # in B     */
	int i, costa, costb;
	int greediness;
	/*int sub_degs[MAX_VERTS];*/

	t_cost=0;			/* keep track of # cuts     */
	t_nongreedy=0;			/* keep track of # ng       */
	t_part_a=0;			/* keep track of # in A     */
	t_part_b=0;			/* keep track of # in B     */

	if (sort_type == SUBGRAPH_MAXDEGREE)
		for(i=0;i<num;i++) subgraph_deg[i] = deg[i];

	if (Output_GP && best)
		init_gp_info();
	num_vertices[SET_A] = num_vertices[SET_B] = 0 ;
	greedy = nongreedy  = 0;

	if (best && (NG_Upflag || Nongreedy_Stats))
		init_ng_placed_neighbors();


	/********************************************************************/
	/**** 2/21/93						     ********/
	/**** set up "pretend" partitioning of unplaced vertices     ********/
	/********************************************************************/
	if (best)
	   for (i=0; i<num; i++)  
	     node_downcost[0][i]=node_downcost[1][i]=part_neigh_diffs[i]=0;
	/********************************************************************/

	for(i=0;i<num;i++) {
		if (best || change_index < num) Tot_Cost[i] = 0;
		if (best || change_index < num) NG_Count[i] = 0;
		if (best || change_index < num) Size_A[i] = 0;
		if (best || change_index < num) Size_B[i] = 0;
		vertex_processed[i] = FALSE ;
		node_cost[SET_A][i] = 0 ;
		node_cost[SET_B][i] = 0 ;
		even_b[i] = FALSE ;
	}
	First = TRUE ; /** first vertex, always choose maximum on equal **/

	for(i=0;i<num;i++) {
		if (duplicate_p_n)
			vertices[i] = get_next_p_n_prime(hold_vertices[i]);
		else if (alg_type == BACKTRACKING ||
			 (alg_type == ALG_TYPE_KL && sort_type == 7)) 
			vertices[i] = get_next_concurrent(best);
		else if (alg_type == ALG_TYPE_KL && sort_type == 10) 
			vertices[i] = get_next_sub_maxdeg(best);
		else if(alg_type==ALG_TYPE_KL && sort_type==8 && i<num-max_size)
			vertices[i] = get_next_concurrent(best);

		/********temporary*****************/
		/*sub_degs[i] = sub_deg(vertices[i]);
		if (i == (num-1)) {
			printf("SUBGRAPH DEGREES:\n");
			printf("_________________\n");
			for (j=0; j<num; j++) 
				printf("%d ", sub_degs[j]);
		}*/
		/*************************/
		   
		get_post_cost(i, vertices[i],&costa,&costb,best,change_index) ;

		/*********************************************************/
		/*if ((i <= begin_fix) || (i >= end_fix)) {*/
			greediness=find_up_greediness(i,vertices[i],best,
							costa,costb);
			find_down_greediness(i, vertices[i], best);
			find_both_greediness(i, vertices[i], best);
		/*}*/
		if (Output_GP && best && greediness==NONGREEDY)
		   gp_info(vertices[i], subgraph_deg[i],
			   vertex_loc[vertices[i]], costa, costb);
		if (vertex_loc[vertices[i]] == SET_A)   {
			t_cost += costa;
			t_part_a += 1;
		} else {
			t_cost += costb;
			t_part_b += 1;
		}
		t_nongreedy += greediness;
		if (best || change_index < num)    Tot_Cost[i] = t_cost;
		if (best || change_index < num)    NG_Count[i] = t_nongreedy;
		if (best || change_index < num)    Size_A[i] = t_part_a;
		if (best || change_index < num)    Size_B[i] = t_part_b;
    	}
	if (best && Output_Gnuplot)
	    	dump_cutfiles(vertices, 0, (int) (num * 0.9));
   /** any even step which was even and still is even, make greedy **/
	for(i=0;i<num;i++) {
	/***********************************************
       		if(even_b[i]) {
       	   		if(node_cost[SET_A][vertices[i]] == 
             			node_cost[SET_B][vertices[i]]) {
               			even_b[i] = FALSE ;
               			greedy++;
               			nongreedy--;
	     		}
		}
	**************************************************/
		if(best) {
         		hold_even_b[i] = even_b[i];
         		if(alg_type == ALG_TYPE_KL) {
            			hold_tie_1[i] = tie1;
            			hold_tie_2[i] = tie2;
	  		}
		}
	}
	/*****************************************************************/
	/** save nongreedy vertices of best cost pass in terms of cuts  **/
	/** (not in terms of greediness/nongreediness)			**/
	/*****************************************************************/
	if (best) {
		for (i=0; i<nongreedy; i++) {
			hold_nongreedy_vertices[i].v = nongreedy_vertices[i].v;
			hold_nongreedy_vertices[i].rank = 
						    nongreedy_vertices[i].rank;
		}
		for (i=0; i<num; i++) {
         		hold_vertices[i] = vertices[i];
         		hold_vertex_loc[vertices[i]] = vertex_loc[vertices[i]];
			hold_post_tots[i] = post_tots[i];
			hold_post_tots_down[i] = post_tots_down[i];
			hold_post_tots_both[i] = post_tots_both[i];
		}
	}
	if (best && !((sort_type == BACKTRACK) && 
		      (nongreedy >= hold_nongreedy))) {
		hold_nongreedy = nongreedy; 
		hold_greedy = greedy; 
	}
	/*****************************************************************/
	if ((change_index<0 || change_index>num-1) && (t_cost != currentcost))
		printf("error 46 in post_proc()\n");
 }

static int intcompare(int *i, int *j)
{
	return(*i - *j);
}

void post_processing(int best, int lower, int upper)
/*  lower:		verts from lower to upper were fixed during */
/*  upper:		the KL run.  Don't count these in the postp.*/
{
	extern int Change_Index;
	extern int hold_loc[MAX_VERTICES];
	extern LISTPTR nodes;
	int partition_diff = num;
	int i ;

#ifdef BUCKET
	for (i=0; i<num; i++)
		vertex_loc[i] = nodes[i].loc;
#endif

	if (sort_type != BACKTRACK) {
		post_proc(best, Change_Index, FALSE, lower, upper);
      		/*hold_num_vertices[SET_A] = num_vertices[SET_A];
      		hold_num_vertices[SET_B] = num_vertices[SET_B];*/
		return;
	}
	init_backtrack();
	greedy_tree_walk(mode, 1, DONT_PRINT, 0);
	/*while (backtrack(mode)  != BOS) { */
		post_proc(best, Change_Index, FALSE, lower, upper);
      		/*hold_num_vertices[SET_A] = num_vertices[SET_A];
      		hold_num_vertices[SET_B] = num_vertices[SET_B];*/
	/*}*/
 	fprintf(outfile,"PARTITION DIFFERENCES:\n");
	for (i=0; i<num; i++) {
		partition_diff-=(vertex_loc[i]==hold_loc[i]);
		if (vertex_loc[i] != hold_loc[i])
			fprintf(outfile, "%4d", i);
	}
 	/*fprintf(outfile,"\nNONGREEDY VERTICES:\n");
	qsort(nongreedy_vertices, nongreedy, sizeof(VLIST), intcompare);
	for (i=0; i<nongreedy; i++) 
		fprintf(outfile, "%4d", nongreedy_vertices[i].v);*/
        fprintf(outfile,"\n%d same, %d different\n",
                         num-partition_diff,partition_diff);
}
/*********************************************************/
/*************  clear summary information  ***************/
/*********************************************************/
void clear_summary()
{  Hnum = VERY_LOW;
   Hprob = 0.0;
   Tcount = 0;
   Tedges = 0;
   Titerations = 0 ;
   Tcuts = 0;
   Tquotient = 0.0;
   Tgreedy = 0 ;
   Tnongreedy = 0 ;
   Ttime = 0.0 ;
   T_ave_cands = 0.0;
   T_leaf_count = 0.0;
   Tlast = 0;                                  
   Tcatch = 0;                                  
   Tsurpass = 0;                                  
   Tchanges = 0;                                  
   T_Ave_Path_Length = 0.0;
   T_Num_Augmenting_Paths = 0.0;
   T_Ave_Flip_Cost = 0.0;
   T_NG_Count = 0;
   T_NG_Before_100 = 0;
   T_NG_Before_250 = 0;
   T_NG_Before_400 = 0;
   Tpart_match = 0;
   Tkl_pg_index = 0;
   Tchange_ave = 0.0;                                  
   Tno_change_ave = 0.0;                                  
 }

/*****************************************************/
/***********  print summary information  *************/
/*****************************************************/
void print_summary()
{
     extern int mode;
     char   solution[20] = "                   ";

     if (mode==MIN_QUOTIENT)
	sprintf(solution,"%.6f & %.6f", Tquotient/(double)Tcount, 
					  (double)Tcuts/(double)Tcount);
     else
	sprintf(solution,"%.6f", (double)Tcuts/(double)Tcount);
     if ((tfile != NULL) && (alg_type == BACKTRACKING) && Paths !=BACKTRACK) {
	 	fprintf(tfile,"SUMMARY:\n");
	 	fprintf(tfile,"%d & %.6f & %.6f & %.6f & %.6f & %20s & %.6f & %.6f & %.6f & %.6f & %.6f & %.6f\n",
	    	Hnum, Hprob, (double) Tedges / (double) Tcount, 
		Titerations / (double) Tcount,
		Ttime/(double) Tcount, 
		solution,
		T_ave_cands / (double) Tcount,
		/*T_leaf_count/ (double) Tcount,*/
		T_NG_Count/ (double) Tcount,
		T_NG_Before_400/ (double) Tcount,
		/*T_NG_Before_250/ (double) Tcount,*/
		T_Ave_Flip_Cost/ (double) Tcount,
		T_Num_Augmenting_Paths/ (double) Tcount,
		T_Ave_Path_Length   / (double) Tcount);
		/*T_NG_Before_100/ (double) Tcount,
		Tkl_pg_index   / (double) Tcount);*/
         	fflush(tfile);
       	 	clear_summary();
		return;
     }
   /** print summary information **/
   if(Tcount > 0 /* && 
         (num != Hnum || prob != Hprob || sort_type != Hsort_type)*/) {
      if(Tcount > 1 && !output_short) {
         fprintf(outfile,"\n");
         fprintf(outfile,"==================================\n");
         fprintf(outfile,"======= Summary Information ======\n");
         fprintf(outfile,"==================================\n");
         fprintf(outfile,"      total passes: %d\n",Tcount);
         fprintf(outfile,"number of vertices: %d\n",Hnum);
         if(Hprob != 0.0)
            fprintf(outfile,"  edge probability: %.6f\n",Hprob);
         fprintf(outfile,"         sort type: ");
         switch(Hsort_type) {
            case 0 : fprintf(outfile,"random\n");
                     break;
            case 1 : fprintf(outfile,"decreasing degree\n");
                     break;
            case 2 : fprintf(outfile,"increasing degree\n");
                     break;
            case 3 : fprintf(outfile,"forward decreasing degree\n");
                     break;
            case 4 : fprintf(outfile,"forward increasing degree\n");
                     break;
            case 5 : fprintf(outfile,"single independent set\n");
                     break;
            case 6 : fprintf(outfile,"multiple independent set\n");
                     break;
            case 7 : fprintf(outfile,"concurrent reordering\n");
                     break;
            case 8 : fprintf(outfile,"ind. set with conc. reordering\n");
                     break;
            case BACKTRACK : fprintf(outfile,"backtracking \n");
                     break;
            default: fprintf(outfile,"*****unknown*****\n");
	    }
         fprintf(outfile," avg. number edges: %6.2f\n",(double) Tedges/ (double) Tcount);
         if(alg_type == ALG_TYPE_KL)
	    fprintf(outfile,"   avg. iterations: %6.2f\n",(double) Titerations/(double) Tcount);
         fprintf(outfile,"         avg. time: %6.2f\n",Ttime / (double) Tcount);
         fprintf(outfile,"         avg. cuts: %6.2f\n",(double) Tcuts / (double) Tcount);
         fprintf(outfile,"       avg. greedy: %6.2f\n",(double) Tgreedy / (double) Tcount);
         fprintf(outfile,"   avg. non-greedy: %6.2f\n",(double) Tnongreedy / (double) Tcount);
         fprintf(outfile,"\n\n");
       }
      /** write summary information to summary file if requested **/
	if(tfile != NULL) {
	 fprintf(tfile,"%d & %d & %.6f & %d & %.6f & %.6f & %.6f & %20s & %.6f & %.6f & %.6f & %.6f & %.6f & %.6f & %.6f %.6f",
	    Tcount, Hnum, Hprob, Hsort_type, (double) Tedges / (double) Tcount,
	    (double) Titerations / (double) Tcount,
	    Ttime/(double) Tcount, 
	    solution,
	    (double) Tgreedy / (double) Tcount,
	    (double) Tnongreedy / (double) Tcount,
	    (double) Tlast / (double) Tcount,
	    (double) Tcatch / (double) Tcount,
	    (double) Tsurpass / (double) Tcount,
	    (double) Tchanges / (double) Tcount,
	    (double) Tpart_match / (double) Tcount,
	    (double) Tkl_pg_index / (double) Tcount);
	    /*(double) Tchange_ave / (double) Tcount,
	    (double) Tno_change_ave / (double) Tcount*/ 

	fprintf(tfile,"\n");
        fflush(tfile);
       }
      /*clear_summary();*/
    }
}

/*******************************************************/
/************* add info to summary files   *************/
/*******************************************************/
void add_to_summary()
{
extern int Nongreedy_Stats;			
extern int Heuristic_Only;		/** count kl runs instead of 	**/
					/** iterations.			**/
extern int Paths;			/* pertains to bk	     	**/
extern int Num_Runs;			/* #times entire alg was run  	**/
extern int iter;			/* #passes in 1 kl run       	**/
extern int Last_Meaningful_Nongreedy;
extern int Catch_Up;
extern int Surpass;
extern int P_Diff;
extern int Changes[];
extern int Neighbors_At_Level[];
extern int Hold_NG_Count;
extern int Hold_NG_Before_100;
extern int Hold_NG_Before_250;
extern int Hold_NG_Before_400;
extern int Part_Match_Point;		/* place where simultaneous greedy */
					/* placement of remaining vertices */
					/* duplicates KL partitioning      */

extern double T_ng_upneighbors,		/* these to keep track of per-pass */
           T_upneighbors,		/* information                	   */
	   T_upneighbors_g,
	   T_upneighbors_n,
           T_ng_upneighbors_g,
           T_ng_up_same_part_g,
           T_ng_up_other_part_g,
           T_ng_upneighbors_n,
           T_ng_up_same_part_n,
           T_ng_up_other_part_n;

extern double Ave_Path_Length;		/* from PO (local_opt.c)	 **/
extern int Num_Augmenting_Paths;	/* from PO (local_opt.c)	 **/
extern int Total_Flip_Cost;		/* from PO (local_opt.c)	 **/

/*extern double T_Ng_upneighbors,	  these were to keep track of summary  
           T_Upneighbors,		  information over all passes	   
	   T_Upneighbors_g,
	   T_Upneighbors_n,
           T_Ng_upneighbors_g,
           T_Ng_up_same_part_g,
           T_Ng_up_other_part_g,
           T_Ng_upneighbors_n,
           T_Ng_up_same_part_n,
           T_Ng_up_other_part_n;*/

extern double Hold_Ave_Cands;
extern double leaf_count;
extern double Change_Ave;
extern double No_Change_Ave;

int i;
char solution[20] = "                   ";

	Hsort_type = sort_type ;
	Hnum = num;
	Hprob = prob;
	Tcount++;
	Tedges = Tedges + numedges;
	if (alg_type == ALG_TYPE_KL) {
		if (Heuristic_Only) {
   			Titerations = Titerations + Num_Runs;
			iter = Num_Runs;
		} else
   			Titerations = Titerations + iter;
	} else
   		Titerations = Titerations + Paths;
	Tkl_pg_index = Tkl_pg_index + KL_PG_Index;
	Tcuts = Tcuts + currentcost;
	Tquotient = currentminsize? 
			Tquotient + ((double) currentcost/(double) currentminsize)
			: 0.0;
	Tgreedy = Tgreedy + hold_greedy;
	Tnongreedy = Tnongreedy + hold_nongreedy;
	Tlast = Tlast + Last_Meaningful_Nongreedy;
	Tcatch = Tcatch + Catch_Up;
	Tsurpass = Tsurpass + Surpass;
	Tchanges = Tchanges + P_Diff;
	Tchange_ave = Tchange_ave + Change_Ave;
	Tno_change_ave = Tno_change_ave + No_Change_Ave;

	T_Ave_Path_Length = T_Ave_Path_Length + Ave_Path_Length;
	T_Num_Augmenting_Paths = (Paths)? T_Num_Augmenting_Paths + 
				Num_Augmenting_Paths/(double)Paths
			        : T_Num_Augmenting_Paths;
	T_Ave_Flip_Cost = Num_Augmenting_Paths? T_Ave_Flip_Cost +
				Total_Flip_Cost/(double)Num_Augmenting_Paths
			  : T_Ave_Flip_Cost;

	T_Ng_upneighbors += T_ng_upneighbors;
	T_Upneighbors += T_upneighbors; 
	T_Upneighbors_g += T_upneighbors_g;     
	T_Ng_upneighbors_g += T_ng_upneighbors_g;     
	if (hold_greedy) {

	T_Ng_up_same_part_g += (T_ng_up_same_part_g / (double) hold_greedy);   
	T_Pct_ng_same_g  +=    (T_pct_ng_same_g / (double) hold_greedy);
	T_Ng_up_other_part_g += (T_ng_up_other_part_g / (double) hold_greedy); 
	T_Pct_ng_other_g  +=    (T_pct_ng_other_g / (double) hold_greedy);

	}
	T_Upneighbors_n += T_upneighbors_n;     
	T_Ng_upneighbors_n += T_ng_upneighbors_n;
	if (hold_nongreedy) {

	T_Ng_up_same_part_n += (T_ng_up_same_part_n / (double) hold_nongreedy);
	T_Pct_ng_same_n  +=    (T_pct_ng_same_n / (double) hold_nongreedy);
	T_Ng_up_other_part_n += (T_ng_up_other_part_n / (double) hold_nongreedy);
	T_Pct_ng_other_n  +=    (T_pct_ng_other_n / (double) hold_nongreedy);

	}

	T_NG_Count = T_NG_Count + Hold_NG_Count;
	T_NG_Before_100 = T_NG_Before_100 + Hold_NG_Before_100;
	T_NG_Before_250 = T_NG_Before_250 + Hold_NG_Before_250;
	T_NG_Before_400 = T_NG_Before_400 + Hold_NG_Before_400;
	Tpart_match = Tpart_match + Part_Match_Point;
	Ttime = Ttime + time_;
     	if (mode == MIN_QUOTIENT)
		sprintf(solution,"%.6f & %d",  
				(double)currentcost/ (double)currentminsize,
				currentcost);
     	else
		sprintf(solution,"%d", currentcost);
	if (alg_type == BACKTRACKING) {
	    if (Paths == BACKTRACK)
		T_ave_cands =   T_ave_cands + 
			pow(10.0,(double)log10((double)leaf_count)/num);
	    else
		T_ave_cands =   T_ave_cands + Hold_Ave_Cands;
   	    T_leaf_count = T_leaf_count + leaf_count;
	    if (tfile != NULL) {
	   	if (Paths == BACKTRACK) {
	 		fprintf(tfile,"%d & %.6f & %d & %.6f & %20s & %.6f & %.6f\n",
	    		Hnum, Hprob, numedges,
			time_, 
			solution, 
			pow(10.0,(double)log10((double)leaf_count)/num),
			leaf_count);
         		fflush(tfile);
			return;
	   	} else {
	 		fprintf(tfile,"%d & %.6f & %d & %d & %.6f & %20s & %.6f & %d & %d & %.6f & %d & %.6f\n",
	    		Hnum, Hprob, numedges, Paths,
			time_, 
			solution, 
			Hold_Ave_Cands,
			/*leaf_count, */
			Hold_NG_Count,
			Hold_NG_Before_400,
			/*Hold_NG_Before_250,*/
			/*Hold_NG_Before_100,*/
			Total_Flip_Cost/(double)Num_Augmenting_Paths,
			Num_Augmenting_Paths,
			Ave_Path_Length);
			/*KL_PG_Index);*/
         		fflush(tfile);
			return;
	   	}
	    }
   	}
	if(alg_type == ALG_TYPE_KL && tfile != NULL) {
	fprintf(tfile,"%d & %.6f & %d & %d & %d & %.6f & %20s & %d & %d "
		      "& %d & %d & %d & %d & %d & %.6f & %.6f",
	    	Hnum, Hprob, Hsort_type, numedges,
	    	iter,
	    	time_, 
	    	solution,
	    	hold_greedy,
	    	hold_nongreedy,
	    	Last_Meaningful_Nongreedy,
	    	Catch_Up,
	    	Surpass,
	    	P_Diff,
		Part_Match_Point,
	    	Change_Ave,
	    	No_Change_Ave);
	if (Nongreedy_Stats) {
	   for (i=1; Neighbors_At_Level[i] > 0; i++)
		fprintf(tfile,"&%d/{\\bf %d}",Changes[i],Neighbors_At_Level[i]);
	   fprintf(tfile," & %d/%d\n", Changes[0] , Neighbors_At_Level[0]);
	} else
	   fprintf(tfile,"\n");
        fflush(tfile);
    }
}
